/*
 * Autor: Burungiu Cristian
 * Proiect: Jocul Snake
 * Fisier sursa:    game_snake.c
 */

#include <avr/io.h>
#include <util/delay.h>
#include <stdlib.h>

#include "u8g.h"

#define F_CPU 16000000

#define LCD_WIDTH  128
#define LCD_HEIGHT 64

#define UP   			0
#define DOWN  			1
#define LEFT  			2
#define RIGHT 			3
#define NONE  			5

#define INITIAL_SIZE 	30
#define MAX_SIZE     	150
#define BLOCK_SIZE 		2

#define BAD_COLLISION	2
#define	OK				0

#define REFRESH_RATE	100
#define NITRO_REFRESH	1

typedef struct {
    uint8_t x;
    uint8_t y;
} Point2D;


uint8_t snake_direction; 				// direction to move - by default, to right
Point2D snake[MAX_SIZE*2];				// snake coordinates
Point2D mancare;
uint32_t head_pos;						// head position in snake vector
uint32_t tail_pos;						// tail position in snake vector
uint32_t snake_size;					// snake length
uint32_t game_delay = REFRESH_RATE;
u8g_t u8g;								// LCD



/* Seteaza directia porturilor*/
void init_buttons(void) {
    DDRD = 0x00; //Seteaza portul D ca input
    PORTD = 0xff; // Activate pull-up resistors on port D   
	
	DDRD |= (1 << PD7);
	PORTD &= ~(1 << PD7);
    
}

/* Verifica miscarea la stanga*/
void check_left(void) {
	if (!(PIND & (1 << PD0)) && snake_direction != DOWN) {
		if (snake_direction == UP) game_delay = NITRO_REFRESH;
        snake_direction = UP;
    }
}

void check_right(void) {
    if (!(PIND & (1 << PD1)) && snake_direction != RIGHT) {
        if (snake_direction == LEFT) game_delay = NITRO_REFRESH;
        snake_direction = LEFT;
    }
}

void checK_up(void) {
     if (!(PIND & (1 << PD4)) && snake_direction != LEFT) {
        if (snake_direction == RIGHT) game_delay = NITRO_REFRESH;
        snake_direction = RIGHT;
    }
}

void check_down(void) {
    if (!(PIND & (1 << PD5)) && snake_direction != UP) {
        if (snake_direction == DOWN) game_delay = NITRO_REFRESH;
        snake_direction = DOWN;
    }
}

/* Check buttons */
void check_input(void) {
	game_delay = REFRESH_RATE;
	check_left();
    check_right();
    checK_up();
    check_down();    
}

//Initializeaza pinii controlerului LCD-ului
void init_lcd(void) {
    MCUCR |= (1 << JTD); 
    MCUCR |= (1 << JTD);
    
    // init LCD - data ports PB0 - PB7, control on PA0 - PA4
    u8g_Init8Bit(&u8g, 
            &u8g_dev_st7920_128x64_4x_8bit,
            PN(1, 0), 
            PN(1, 1), 
            PN(1, 2),
            PN(1, 3), 
            PN(1, 4), 
            PN(1, 5), 
            PN(1, 6),
            PN(1, 7), 	// Pini de date
            PN(0, 2),   // Punctul enable
            U8G_PIN_NONE,
            U8G_PIN_NONE,
            PN(0, 0),    // Firul RS
            PN(0, 1),    // Firul RW
            PN(0, 3));   // Firul RST
}

Point2D genereazaMancare(void) {
    Point2D point;
    
    point.x = rand() % LCD_WIDTH;
    point.y = rand() % LCD_HEIGHT;
    
    return point;
}

void initializeSnake(uint32_t seed) {
    uint8_t i;
    
    snake_size = INITIAL_SIZE;
    
    for (i = 0; i < INITIAL_SIZE; ++i) {
        snake[i].x = (LCD_WIDTH - INITIAL_SIZE) / 2 + i * BLOCK_SIZE;
        snake[i].y = LCD_HEIGHT / 2;
    }

    head_pos = INITIAL_SIZE - 1;
    tail_pos = 0;


    srand(seed);
    
    // genereaza mancare
    mancare = genereazaMancare();
}

void makeDelay(void) {
	for (int i = 0; i < 500; i++) {
		PORTD |= (1 << PD7);
		_delay_ms(0.1);
		PORTD &= ~(1 << PD7);
		_delay_ms(0.1);
	}
}

void makeLongerDelay(void) {
	for (int i = 0; i < 4000; i++) {
		PORTD |= (1 << PD7);
		_delay_ms(0.1);
		PORTD &= ~(1 << PD7);
		_delay_ms(0.3);
	}
}

void checkSakeDirection(void) {
    switch(snake_direction) {
        case UP:
            snake[head_pos].y = (snake[head_pos].y + BLOCK_SIZE) % LCD_HEIGHT;
            break;
        case DOWN:
            snake[head_pos].y = (snake[head_pos].y - BLOCK_SIZE) % LCD_HEIGHT;
            break;
        case LEFT:
            snake[head_pos].x = (snake[head_pos].x - BLOCK_SIZE) % LCD_WIDTH;
            break;
        case RIGHT:
            snake[head_pos].x = (snake[head_pos].x + BLOCK_SIZE) % LCD_WIDTH;
            break;
    }
}

int checkCollision(void) {
    int i;
    for (i = 0; i < snake_size; ++i) {
        Point2D block = snake[(tail_pos + i) % MAX_SIZE];
        if (abs(block.x - snake[head_pos].x) < BLOCK_SIZE 
                && abs(block.y - snake[head_pos].y) < BLOCK_SIZE)
            return 1;
    }
    return 0;
}


uint8_t checkEat(void) {
    uint8_t status;
    if (abs(mancare.x - snake[head_pos].x) < BLOCK_SIZE && 
        abs(mancare.y - snake[head_pos].y) < BLOCK_SIZE) {
        ++snake_size;
        mancare = genereazaMancare();
        status = OK;
        makeDelay();
    } else {
        tail_pos = (tail_pos + 1) % MAX_SIZE;
        status = OK;
    }
    return status;
}

uint8_t update_game(void) {
    uint32_t prev;
    uint8_t status;

    prev = head_pos;
    head_pos = (head_pos + 1) % MAX_SIZE;
    
    snake[head_pos] = snake[prev];
    
    checkSakeDirection();

	if (checkCollision())
        return BAD_COLLISION;
    status = checkEat();
	
	return status;
}

void draw_start(void) {
    u8g_SetDefaultForegroundColor(&u8g);
    u8g_SetFont(&u8g, u8g_font_helvB10);
    u8g_DrawStr(&u8g, 5, 25, "!!Snake Game!!");
	u8g_DrawStr(&u8g, 5, 40, "Press down key!");
	u8g_DrawStr(&u8g, 5, 55, "Proiect PM");
}

void draw_game(void) {
    uint32_t i;
    Point2D block;
    
    u8g_SetDefaultForegroundColor(&u8g);
    for (i = 0; i < snake_size; ++i) {
        block = snake[(tail_pos + i) % MAX_SIZE];
        u8g_DrawBox(&u8g, block.x - BLOCK_SIZE / 2, block.y - BLOCK_SIZE / 2, BLOCK_SIZE, BLOCK_SIZE);
    }
    
    // Deseneaza mancare
    u8g_DrawBox(&u8g, mancare.x - BLOCK_SIZE / 2, mancare.y - BLOCK_SIZE / 2, BLOCK_SIZE, BLOCK_SIZE);
}

int main(void) {
	uint32_t seed = 42;
	uint8_t status;
	
begin:	

    init_buttons();
    init_lcd();
    
    snake_direction = NONE;
    
    // Se asteapta pana se introduce prima comanda
    while(1) {
        check_input();
        
        // Incepe jocul
        if (snake_direction != NONE) break;
        
        u8g_FirstPage(&u8g);
        do {
            draw_start();
        } while (u8g_NextPage(&u8g));    
        
		seed += seed * 8;
        _delay_ms(REFRESH_RATE);
    }
    
    initializeSnake(seed);

    while (1) {
	
        check_input(); 
		status = update_game();
		
        if (status == BAD_COLLISION) {
			makeLongerDelay();
			goto begin;
		}
			
        u8g_FirstPage(&u8g);
        do {
            draw_game();
        } while (u8g_NextPage(&u8g)); 
		
		if (game_delay == REFRESH_RATE)
			_delay_ms(game_delay);
    }
    
    return 0;
}